import logging
import os
from typing import Any, Callable, Dict

from mitogen.core import Error, StreamError
from mitogen.parent import Router
from mitogen.utils import run_with_router

from . import core
from .config import create_data, get_config, path_to_config_file, set_config, set_data
from .deps import Modules, TransmitModules, makereal, maketransmit, runfunc


def decode(info):
    for k in list(info.keys()):
        if type(k) == bytes:
            info[k.decode()] = info[k].decode()
            del info[k]


def main(router: Router, func: Callable[..., None], *args: Any, **kwargs: Any) -> Dict:
    config = get_config()
    calls = []
    wg = core.is_wireguard()
    data = None
    for server in config["servers"]:
        assert isinstance(server, Dict)
        hostname = server["wireguard_ip"] if wg else server["ssh_hostname"]
        port = 22 if wg else server.get("ssh_port", 22)
        key_path = path_to_config_file(server["ssh_key"])
        if not os.path.exists(key_path):
            raise Exception(f"Can't find ssh key {key_path}")
        try:
            connect = router.ssh(
                hostname=hostname,
                port=port,
                username=server["ssh_user"],
                identity_file=key_path,
                check_host_keys="accept",
                python_path="python3",
            )
        except StreamError:
            print(
                "Exception while trying to login to %s@%s:%s"
                % (server["ssh_user"], hostname, port)
            )
            raise

        sudo = router.sudo(via=connect, python_path="python3")
        data = create_data(server=server)
        calls.append(sudo.call_async(func, data, *args, **kwargs))

    infos = []
    errors = []
    for call in calls:
        try:
            info = call.get().unpickle()
            if info is not None:
                decode(info)
            infos.append(info)
        except Error as e:
            print("Got error", e)
            errors.append(e)

    if len(errors) > 0:
        raise Exception(errors)

    return {"infos": infos, "data": data}


def do(data, transmitmodules: TransmitModules, name: str):
    set_data(data)
    modules = makereal(transmitmodules)
    return runfunc(modules, name)


def internal_runner(
    router: Router, modules: Modules, local_func: str, run_func: str, parse_func: str
) -> None:
    runfunc(modules, local_func)
    infos = main(router, do, maketransmit(modules), run_func)
    for info in infos["infos"]:
        runfunc(modules, parse_func, info, infos["data"])


def run(inventory_path: str, modules: Modules):
    logging.basicConfig()
    logging.root.setLevel(logging.INFO)
    set_config(inventory_path)

    modules.insert(0, core)
    needs_dependencies = list(modules)
    while len(needs_dependencies) > 0:
        new_dependencies = runfunc(needs_dependencies, "dependencies")
        needs_dependencies = []
        for item in new_dependencies.values():
            for new_dependency in item[0]:
                if new_dependency in modules:
                    continue
                modules.insert(0, new_dependency)
                needs_dependencies.append(new_dependency)

    print("Running:")
    for module in maketransmit(modules):
        print(f"* {module}")
    print("")

    run_with_router(
        internal_runner,
        modules,
        "bootstrap_local",
        "bootstrap_run",
        "bootstrap_parse_return",
    )
    run_with_router(
        internal_runner, modules, "core_local", "core_run", "core_parse_return"
    )



import json
import os
import pathlib

from .config import config_path, get_config_file
from .fs import run_command


def bootstrap_local():
    aws_path = pathlib.Path(config_path()).joinpath("other-aws")
    if not aws_path.exists():
        json.dump(
            {
                "access_key": run_command(
                    "aws configure get aws_access_key_id"
                ).strip(),
                "secret_key": run_command(
                    "aws configure get aws_secret_access_key"
                ).strip(),
            },
            aws_path.open("w"),
            indent=2,
        )


def set_aws_creds():
    aws_conf = json.loads(get_config_file("configs/other-aws"))
    os.environ["AWS_ACCESS_KEY_ID"] = aws_conf["access_key"]
    os.environ["AWS_SECRET_ACCESS_KEY"] = aws_conf["secret_key"]



import json
import os
import socket
from pathlib import Path
from typing import Dict, List

from .config import config, host, network_config_file, other_config_file
from .debian import apt_install
from .fs import run_command
from .users import users


def is_wireguard():
    return os.path.exists("/etc/wireguard")


def hash_fn(key: str, count: int) -> int:
    return sum(bytearray(key.encode("utf-8"))) % count


# Use this host for a given service
# Intended for "run on one machine" things
def use_this_host(name: str) -> bool:
    hosts = [h["name"] for h in config()["servers"]]
    index = hash_fn(name, len(hosts))
    return host()["name"] == hosts[index]


def bootstrap_run():
    apt_install(["iproute2"])

    data = {
        "hostname": socket.gethostname(),
        "network_devices": run_command("ip -j address"),
        "users": users(force_load=True),
        "groups": run_command("getent group"),
        "server_name": host()["name"],
    }
    ip_file = Path("/opt/ip_address")
    if ip_file.exists():
        data["external_ip"] = json.load(ip_file.open())
    else:
        if in_vagrant() or in_docker():
            networks = json.loads(data["network_devices"])
            ext_if = [net for net in networks if net["ifname"] == "eth0"]
            if len(ext_if) > 0 and len(ext_if[0]["addr_info"]) > 0:
                data["external_ip"] = json.dumps(
                    {"ip": ext_if[0]["addr_info"][0]["local"]}
                )
            else:
                data["external_ip"] = "<unknown>"
        else:
            apt_install(["curl", "ca-certificates"])
            data["external_ip"] = run_command("curl https://api.ipify.org?format=json")
        json.dump(data["external_ip"], ip_file.open("w"))

    return data


def bootstrap_parse_return(infos: List[Dict]) -> None:
    info = infos[0]
    networks = json.loads(info["network_devices"])
    name = info["server_name"]
    json.dump(networks, open(network_config_file(name), "w"), indent=2)

    other = {
        "external_ip": json.loads(info["external_ip"])["ip"],
        "users": info["users"],
        "groups": info["groups"],
        "hostname": info["hostname"],
    }
    json.dump(other, open(other_config_file(name), "w"), indent=2)


def in_vagrant():
    return "vagrant" in users()


def in_docker():
    return os.path.exists("/.dockerenv")



import logging
import subprocess
from pathlib import Path

from .fs import link, run_command


def journal(name):
    res = run_command("journalctl -u %s" % name)
    print(res)


def systemd_set(name, enabled=None, running=None, restart=None, reloaded=None):
    raw = run_command("systemctl show %s --no-page" % name)
    status = dict([line.split("=", 1) for line in raw.splitlines()])
    if status.get("UnitFileState") == "masked":
        logging.info("Unmasking %s" % name)
        run_command("systemctl unmask %s" % name)
    if enabled is not None:
        if enabled:
            if status["UnitFileState"] not in ["enabled", "enabled-runtime"]:
                logging.info("%s is currently %s" % (name, status["UnitFileState"]))
                run_command("systemctl enable %s" % name)
        else:
            if status["UnitFileState"] != "disabled":
                logging.info("%s is currently %s" % (name, status["UnitFileState"]))
                run_command("systemctl disable %s" % name)
    started = False
    if running is not None:
        if running:
            if status["SubState"] not in ["running", "auto-restart", "start"]:
                logging.info("running: %s %s" % (name, status["SubState"]))
                try:
                    run_command("systemctl start %s" % name)
                except subprocess.CalledProcessError:
                    journal(name)
                    raise
                started = True
        else:
            if status["SubState"] not in ["dead"]:
                logging.info("running: %s %s" % (name, status["SubState"]))
                try:
                    run_command("systemctl stop %s" % name)
                except subprocess.CalledProcessError:
                    journal(name)
                    raise
    if restart is True and not started:
        try:
            run_command("systemctl restart %s" % name)
        except subprocess.CalledProcessError:
            journal(name)
            raise

    if reloaded is True:
        run_command("systemctl reload %s" % name)


def systemctl_daemon_reload():
    run_command("systemctl daemon-reload")


def link_service(fullpath: str) -> bool:
    path = Path(fullpath)
    link_change = link(
        f"/etc/systemd/system/{path.name}",
        path,
    )
    if link_change:
        systemctl_daemon_reload()
    return link_change



import json
import pathlib
from typing import Optional

from .config import config_path, get_config_file
from .fs import set_file_contents_from_template

mailto: Optional[str] = None
mailfrom: Optional[str] = None


def bootstrap_local():
    global mailto, mailfrom
    cron_path = pathlib.Path(config_path()).joinpath("cron-info")
    if not cron_path.exists():
        json.dump(
            {"mailto": mailto, "mailfrom": mailfrom},
            cron_path.open("w"),
            indent=2,
        )


def set_mailto(email: str) -> None:
    global mailto
    mailto = email


def set_mailfrom(email: str) -> None:
    global mailfrom
    mailfrom = email


def create_cron(name: str, schedule: str, user: str, command: str):
    cron_info = json.loads(get_config_file("configs/cron-info"))
    envs = {}
    if cron_info["mailto"] is not None:
        envs["MAILTO"] = cron_info["mailto"]
    if cron_info["mailfrom"] is not None:
        envs["MAILFROM"] = cron_info["mailfrom"]
    set_file_contents_from_template(
        f"/etc/cron.d/{name}",
        "cron.j2",
        SCHEDULE=schedule,
        USER=user,
        COMMAND=command,
        ENVS=envs,
    )



import os
import re
from typing import Dict, List, Optional, Union

from debian.debian_support import version_compare

from .fs import build_with_command, download, run_command, set_file_contents

host_arch: Optional[str] = None

_version_pattern = re.compile(r"Version: (\S+)")


def apt_update():
    run_command("apt-get update --allow-releaseinfo-change")


def add_trusted_key(url: str, name: str, hash: str):
    download(
        url,
        f"/etc/apt/trusted.gpg.d/{name}.gpg.asc",
        hash,
    )
    apt_install(["gpg"])
    build_with_command(
        f"/etc/apt/trusted.gpg.d/{name}.gpg",
        f"cat /etc/apt/trusted.gpg.d/{name}.gpg.asc | gpg --dearmor > /etc/apt/trusted.gpg.d/{name}.gpg",
        [f"/etc/apt/trusted.gpg.d/{name}.gpg.asc"],
    )
    return f"/etc/apt/trusted.gpg.d/{name}.gpg"


# List is just "any version", Dict is a "name => min version" requirement
def apt_install(
    packages: Union[List[str], Dict[str, Optional[str]]],
    always_install: bool = False,
    target_release: Optional[str] = None,
) -> bool:
    global host_arch
    if host_arch is None and packages != ["dpkg-dev"]:
        apt_install(["dpkg-dev"])
        host_arch = run_command("dpkg-architecture -q DEB_HOST_ARCH").strip()
    if isinstance(packages, List):
        packages = dict([(p, None) for p in packages])
    if always_install:
        to_install = list(packages.keys())
    else:
        to_install = []
        for package in packages.keys():
            paths = [
                f"/var/lib/dpkg/info/{package}.list",
                f"/var/lib/dpkg/info/{package}:{host_arch}.list",
            ]
            for path in paths:
                if not os.path.exists(path):
                    continue
                wanted_version = packages[package]
                if wanted_version is None:  # existance is enough
                    break
                status = run_command(f"dpkg-query --status {package}")
                version = _version_pattern.search(status).group(1)
                if version_compare(version, wanted_version) >= 0:
                    break
            else:
                to_install.append(package)

        if to_install == []:
            return False

    # Because it's probably the first run
    if packages == {"dpkg-dev": None}:
        apt_update()

    # Confdef is to fix https://unix.stackexchange.com/a/416816/73838
    os.environ["DEBIAN_FRONTEND"] = "noninteractive"
    cmd = (
        "apt-get install %s --no-install-recommends --yes -o DPkg::Options::=--force-confdef"
        % " ".join(to_install)
    )
    if target_release is not None:
        cmd += f" --target-release {target_release}"
    run_command(cmd)
    return True


def debian_repo(name, contents=None):
    fname = "/etc/apt/sources.list.d/%s.list" % name
    if contents is None:
        contents = "deb http://deb.debian.org/debian %s main" % name

    changed = set_file_contents(fname, contents)
    if changed:
        apt_update()


def set_alternative(alt_name: str, option: str):
    current_alt_path = os.path.realpath(f"/etc/alternatives/{alt_name}")
    if current_alt_path != option:
        print(
            f"current alt path for {alt_name} is {current_alt_path} not {option}, so fixing"
        )
        run_command(f"update-alternatives --set {alt_name} {option}")



from datetime import timedelta
from pathlib import Path
from typing import Dict, List, Union

from . import aws, cron
from .config import core_config, other_config_file
from .core import use_this_host
from .debian import apt_install
from .deps import Modules
from .fs import (
    make_directory,
    run_command,
    run_with_marker,
    set_file_contents_from_template,
)
from .python import setup_venv

options = {}


# Are we in a test config where we should just not get the cert
def get_dummy_certs():
    config = core_config()
    return config.get("dummy_certs", False)


def certbot_for_host(hostname: Union[str, List[str]], email: str) -> Dict:
    if type(hostname) == str:
        hostnames = [hostname]
    else:
        hostnames = hostname
    certbot = Path("/opt/certbot")
    live_path = certbot.joinpath("config", "live", hostnames[0])

    dummy_certs = get_dummy_certs()

    if use_this_host("certbot"):
        aws.set_aws_creds()
        venv = certbot.joinpath("venv")
        venv_bin = venv.joinpath("bin")
        pip = venv_bin.joinpath("pip")
        certbot_bin = venv_bin.joinpath("certbot")

        fullchain_path = live_path.joinpath("fullchain.pem")
        renew_command = f"{certbot_bin} renew \
                        --config-dir={certbot.joinpath('config')} \
                        --work-dir={certbot.joinpath('workdir')} \
                        --logs-dir={certbot.joinpath('logs')} \
                        --dns-route53"
        while renew_command.find("  ") != -1:
            renew_command = renew_command.replace("  ", " ")
        if not fullchain_path.exists():
            make_directory(live_path)
            make_directory(certbot)
            setup_venv(venv)
            set_file_contents_from_template(
                "/opt/certbot/requirements.txt", "certbot_requirements.txt"
            )
            run_with_marker(
                "/opt/certbot/deps_installed",
                f"{pip} install -r /opt/certbot/requirements.txt",
            )

            if dummy_certs:
                fullchain_path.open("w").write("")
                live_path.joinpath("privkey.pem").open("w").write("")
            else:
                run_command(
                    f"{certbot_bin} certonly \
                        --config-dir={certbot.joinpath('config')} \
                        --work-dir={certbot.joinpath('workdir')} \
                        --logs-dir={certbot.joinpath('logs')} \
                        -m {email} --agree-tos --non-interactive \
                        --no-eff-email --domains {','.join(hostnames)} --dns-route53"
                )
        else:
            if not dummy_certs:
                run_with_marker(
                    "/opt/certbot/renew_marker",
                    renew_command,
                    max_age=timedelta(days=1),
                )

        apt_install(["moreutils"])
        cron.create_cron(
            f"certs-renew-{hostnames[0]}",
            "0 3 * * *",
            "root",
            f"chronic {renew_command}",
        )

        return {
            f"fullchain-{hostnames[0]}": fullchain_path.open().read(),
            f"privkey-{hostnames[0]}": live_path.joinpath("privkey.pem").open().read(),
            "ssl-options": venv.joinpath(
                "lib/python3.9/site-packages/certbot_nginx/_internal/tls_configs/options-ssl-nginx.conf"  # noqa: E501
            )
            .open()
            .read(),
        }
    else:
        return {}


def dependencies() -> Modules:
    return [aws, cron]


def bootstrap_run() -> Dict:
    return certbot_for_host(options["hostname"], options["email"])


def bootstrap_parse_return(
    infos: List[Dict],
) -> None:
    for info in infos:
        for key in info:
            open(other_config_file(key), "w").write(info[key])



import contextlib
import grp
import hashlib
import logging
import os
import pwd
import re
import stat
import subprocess
from datetime import datetime
from difflib import unified_diff
from typing import List, Optional, Union

from .config import data_files, jinja_env


def hash_data(data: bytes) -> str:
    m = hashlib.sha256()
    m.update(data)
    return m.hexdigest()


def set_file_contents(
    fname: str,
    contents: Union[str, bytes],
    ignore_changes: bool = False,
    owner: Optional[str] = None,
    group: Optional[str] = None,
) -> bool:
    needs_update = False

    if type(contents) == bytes:
        try:
            contents = contents.decode("utf-8")
        except UnicodeDecodeError:
            pass

    if not os.path.exists(fname):
        needs_update = True
        logging.info("File %s was missing" % fname)
    elif not ignore_changes:
        if type(contents) == str:
            data = open(fname, "rb").read().decode("utf-8").splitlines(True)
            diff = list(unified_diff(data, contents.splitlines(True)))
            if len(diff) > 0:
                diff = "".join(diff)
                logging.info("File %s was different. Diff is: \n%s" % (fname, diff))
                needs_update = True
        else:
            data = open(fname, "rb").read()
            if hash_data(data) != hash_data(contents):
                logging.info("File %s was different" % fname)
                needs_update = True

    if needs_update:
        if type(contents) == str:
            open(fname, "w").write(contents)
        else:
            open(fname, "wb").write(contents)

    needs_update = set_owner(fname, owner, group) or needs_update

    return needs_update


def set_file_contents_from_template(fname, template, ignore_changes=False, **kwargs):
    return set_file_contents(
        fname,
        jinja_env().get_template(template).render(**kwargs),
        ignore_changes=ignore_changes,
    )


def set_file_contents_from_data(fname: str, data_path: str):
    return set_file_contents(fname, data_files()[data_path])


@contextlib.contextmanager
def cd(path: os.PathLike):
    CWD = os.getcwd()

    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(CWD)


def set_mode(path, mode):
    if type(mode) == str:
        raw_mode = int(mode, 8)
    else:
        raw_mode = mode
    existing = stat.S_IMODE(os.stat(path).st_mode)
    if existing != raw_mode:
        logging.info("chmod %s %s" % (path, mode))
        os.chmod(path, raw_mode)
        return True
    else:
        return False


def set_owner(path, owner: Optional[str] = None, group: Optional[str] = None):
    st = os.stat(path)
    if owner is not None:
        owner_id = pwd.getpwnam(owner).pw_uid
    else:
        owner_id = st.st_uid
    if group is not None:
        group_id = grp.getgrnam(group).gr_gid
    else:
        group_id = st.st_gid

    if st.st_uid != owner_id or st.st_gid != group_id:
        os.chown(path, owner_id, group_id)
        return True
    else:
        return False


def make_directory(path, mode=None, owner=None, group=None):
    ret = False
    if not os.path.exists(path):
        logging.info("Make directory %s" % path)
        os.makedirs(path)
        ret = True
    if mode is not None:
        ret = set_mode(path, mode) or ret
    if owner is not None or group is not None:
        set_owner(path, owner, group)
    return ret


def replace_line(fname, search, replace):
    existing = open(fname).read()
    if search in existing:
        set_file_contents(fname, existing.replace(search, replace))


def insert_line(fname, line):
    existing = open(fname).read()
    if line not in existing:
        return set_file_contents(fname, existing + "\n" + line)
    else:
        return False


def insert_or_replace(fname: str, matcher: re.Pattern, line: str) -> None:
    existing = open(fname).read()
    results = matcher.search(existing)
    if results is not None:
        set_file_contents(
            fname, existing[: results.start()] + line + existing[results.end() :]
        )
    else:
        set_file_contents(fname, existing + "\n" + line)


def sha_file(fname):
    from .debian import apt_install

    apt_install(["coreutils"])
    existing_sha = run_command("sha256sum %s" % fname).strip()
    return existing_sha.split(" ")[0]


def has_sha(fname, sha):
    if os.path.exists(fname):
        existing_sha = sha_file(fname)
        if existing_sha == sha:
            return True

    return False


def download(url, fname, sha, mode=None):
    exists = has_sha(fname, sha)
    if not exists:
        from .debian import apt_install

        apt_install(["curl", "ca-certificates"])
        run_command("curl -Lo %s %s" % (fname, url))
        existing_sha = sha_file(fname)
        assert existing_sha == sha, (existing_sha, sha)

    if mode is not None:
        set_mode(fname, mode)

    return not exists


def link(target, source):
    if os.path.lexists(target) and (
        not os.path.exists(target) or not os.path.samefile(source, target)
    ):
        logging.info("Unlink %s" % target)
        os.remove(target)
    if not os.path.lexists(target):
        logging.info("Link %s to %s" % (target, source))
        os.symlink(source, target)
        return True
    else:
        return False


def download_executable(url, hash, name=None, path=None):
    if name is None:
        name = url.split("/")[-1]
    if path is None:
        path = "/usr/local/bin/%s" % name
    download(
        url,
        path,
        hash,
        mode="755",
    )


def download_and_unpack(url, hash, name=None, dir_name=None, compressed_root="/opt"):
    if name is None:
        name = url.split("/")[-1]
    compressed_path: str = "%s/%s" % (compressed_root, name)
    if dir_name is None:
        dir_name = "/opt/%s" % name.replace(".tar.gz", "").replace(".tgz", "")
    changed = download(
        url,
        compressed_path,
        hash,
    )

    make_directory(dir_name)
    if os.listdir(dir_name) == []:
        from .debian import apt_install

        if compressed_path.endswith("tar.gz") or compressed_path.endswith(".tgz"):
            apt_install(["tar"])
            run_command("tar --directory=%s -zxvf %s" % (dir_name, compressed_path))
        elif compressed_path.endswith("zip"):
            apt_install(["unzip"])
            run_command("unzip %s -d %s" % (compressed_path, dir_name))
        else:
            raise Exception(compressed_path)

        changed = True

    return {"changed": changed, "dir_name": dir_name}


def last_modified(fname):
    try:
        return os.stat(fname).st_mtime
    except FileNotFoundError:
        return float(0)


def delete(fname):
    if os.path.exists(fname):
        logging.info("Deleting %s", fname)
        os.remove(fname)
        return True
    else:
        return False


def build_with_command(fname, command, deps=[], force_build=False, directory=None):
    display = command.strip()
    while display.find("  ") != -1:
        display = display.replace("  ", " ")
    changed = set_file_contents("%s.command" % fname, display) or force_build
    target_modified = last_modified(fname)
    for dep in deps:
        if last_modified(dep) > target_modified:
            logging.info("%s is younger than %s" % (dep, fname))
            changed = True
    if not os.path.exists(fname) or changed:
        logging.info("Building %s" % fname)
        if command.find("|") != -1:
            out = ""
            commands = command.split("|")
            for command in commands:
                try:
                    out = run_command(command, input=out, directory=directory)
                except subprocess.CalledProcessError:
                    print("Ran '%s' with input '%s'" % (command, out))
                    raise
        else:
            run_command(command, directory=directory)
        return True
    else:
        return False


def run_with_marker(
    fname, command, deps=[], max_age=None, force_build=False, directory=None
):
    changed = not os.path.exists(fname) or force_build
    target_modified = last_modified(fname)
    if max_age is not None:
        age = datetime.now() - datetime.fromtimestamp(target_modified)
        if age > max_age:
            changed = True
    for dep in deps:
        dep_modified = last_modified(dep)
        if dep_modified > target_modified:
            logging.info("%s is younger than %s" % (dep, fname))
            changed = True

    if changed:
        run_command(command, directory=directory)
        open(fname, "w").write(command)

    return changed


def run_command(
    cmd: str,
    directory: Optional[str] = None,
    input: Optional[str] = None,
    allowed_exit_codes: List[int] = [0],
) -> str:
    display = cmd.strip()
    while display.find("  ") != -1:
        display = display.replace("  ", " ")
    try:
        if directory is not None:
            logging.info("Run in %s: %s" % (directory, display))
            with cd(directory):
                process = subprocess.Popen(
                    cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    shell=True,
                    stdin=subprocess.PIPE,
                )
        else:
            logging.info("Run: %s" % display)
            process = subprocess.Popen(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=True,
                stdin=subprocess.PIPE,
            )

        if input is not None:
            input = input.encode("utf-8")
        (stdout, stderr) = process.communicate(input=input)
        assert process.returncode in allowed_exit_codes, (
            process.returncode,
            stdout,
            stderr,
        )
        return stdout.decode("utf-8")
    except subprocess.CalledProcessError as e:
        print(e.output)
        raise



from pathlib import Path

from paracrine.config import config_path, data_path, get_config_file, get_config_keys
from paracrine.core import use_this_host
from paracrine.fs import (
    build_with_command,
    download_and_unpack,
    make_directory,
    run_with_marker,
    set_file_contents,
    set_file_contents_from_template,
    set_mode,
)
from paracrine.network import wireguard_ip, wireguard_ips
from paracrine.services import ntp, wireguard
from paracrine.systemd import systemctl_daemon_reload, systemd_set
from paracrine.users import adduser

# https://www.cockroachlabs.com/docs/v22.2/secure-a-cluster


def dependencies():
    return [ntp, wireguard]


cockroach_url = "https://binaries.cockroachdb.com/cockroach-v22.2.9.linux-amd64.tgz"
cockroach_hash = "7ed169bf5f1f27bd49ab4e04a00068f7b44cff8a0672778b0f67d87ece3de07b"
binary_path = "cockroach-v22.2.9.linux-amd64/cockroach"
HOME_DIR = Path("/var/lib/cockroach")
CERTS_DIR = HOME_DIR.joinpath("certs")
USER = "cockroach"


def bootstrap_local():
    store_path = Path(data_path()).joinpath("cockroach")
    make_directory(store_path)
    download_and_unpack(
        cockroach_url,
        cockroach_hash,
        dir_name=store_path,
        compressed_root=data_path(),
    )
    cockroach = store_path.joinpath(binary_path)
    certs_dir = Path(config_path()).joinpath("cockroach-certs")
    make_directory(certs_dir)
    build_with_command(
        store_path.joinpath("ca.key"),
        f"{cockroach} cert create-ca --certs-dir={certs_dir} --ca-key={store_path}/ca.key",
    )
    build_with_command(
        certs_dir.joinpath("client.root.key"),
        f"{cockroach} cert create-client root --certs-dir={certs_dir} --ca-key={store_path}/ca.key",
    )

    for ip in wireguard_ips().values():
        crt_path = certs_dir.joinpath(f"{ip}.crt")
        build_with_command(
            crt_path,
            f"{cockroach} cert create-node localhost {ip} --certs-dir={certs_dir} --ca-key={store_path}/ca.key && mv {certs_dir.joinpath('node.crt')} {crt_path} && mv {certs_dir.joinpath('node.key')} {crt_path.with_suffix('.key')}",
        )


def bootstrap_run():
    unpacked = download_and_unpack(
        cockroach_url,
        cockroach_hash,
    )
    cockroach_path = Path(unpacked["dir_name"]).joinpath(binary_path)
    adduser(USER, HOME_DIR)
    make_directory(CERTS_DIR)
    for fname in get_config_keys():
        if "cockroach-certs" not in fname:
            continue
        new_fname = CERTS_DIR.joinpath(
            fname.replace("configs/cockroach-certs/", "")
        ).as_posix()
        if wireguard_ip() in fname:
            new_fname = new_fname.replace(wireguard_ip(), "node")
        set_file_contents(new_fname, get_config_file(fname), owner="cockroach")
        set_mode(new_fname, "700")
    changes = set_file_contents_from_template(
        "/etc/systemd/system/cockroach.service",
        "cockroach.service.j2",
        COCKROACH_PATH=cockroach_path,
        HOME_DIR=HOME_DIR,
        USER=USER,
        CERTS_DIR=CERTS_DIR,
        WIREGUARD_IP=wireguard_ip(),
        HOST_LIST=",".join([f"{ip}:26257" for ip in wireguard_ips().values()]),
    )
    if changes:
        systemctl_daemon_reload()
    systemd_set("cockroach", enabled=True, running=True, restart=changes)


def core_run():
    if use_this_host("cockroach"):
        run_with_marker(
            HOME_DIR.joinpath("init_done"),
            f"/opt/cockroach-v22.2.9.linux-amd64/cockroach-v22.2.9.linux-amd64/cockroach init --certs-dir={CERTS_DIR} --host={wireguard_ip()}:26257",
        )
